# -*- coding: utf-8 -*-
# (c) 2018-2023, The Earth Observations Developers
from __future__ import print_function
import attr
import json
import logging
import pandas as pd
import pkg_resources
from datetime import datetime
from phenodata.util import haversine_distance, iterate_with_progressbar

logger = logging.getLogger(__name__)


@attr.s
class DwdPhenoData:
    """
    Conveniently access phenology information from the Climate Data Center (CDC)
    FTP server operated by »Deutscher Wetterdienst« (DWD).

    See also `cdc-readme`_:
    Phenological data is collected at about 1200 active stations. The state of development of
    selected plants (e.g., apple, birch, snow drops, goose berry, wheat, vine, etc.) is reported
    by annual reporters and immediate reporters.

    The lists of phenological stations can be found in the list of phenology `annual reporters`_
    and the list of phenology `immediate reporters`_.

    .. _cdc-readme: ftp://opendata.dwd.de/climate_environment/CDC/Readme_intro_CDC_ftp.txt
    .. _annual reporters: ftp://opendata.dwd.de/climate_environment/CDC/help/PH_Beschreibung_Phaenologie_Stationen_Jahresmelder.txt
    .. _immediate reporters: ftp://opendata.dwd.de/climate_environment/CDC/help/PH_Beschreibung_Phaenologie_Stationen_Sofortmelder.txt

    """

    # Instance of the lowlevel DWD CDC FTP server client wrapper object ``phenodata.dwd.cdc.DwdCdcClient``
    cdc = attr.ib()

    # Instance of ``phenodata.dwd.pheno.DwdPhenoDataHumanizer``
    humanizer = attr.ib()

    # The dataset to access, either "annual" or "immediate"
    dataset = attr.ib()

    @property
    def data_directory(self):
        """
        Location of observations on the FTP server
        """
        return '/observations_germany/phenology/{dataset}_reporters'.format(dataset=self.dataset)

    def get_species(self):
        """
        Return DataFrame with species information
        """
        return self.cdc.get_dataframe(path='/help/PH_Beschreibung_Pflanze.txt', index_column=0)

    def get_phases(self):
        """
        Return DataFrame with phases information
        """
        return self.cdc.get_dataframe(path='/help/PH_Beschreibung_Phase.txt', index_column=0)

    def get_quality_levels(self):
        """
        Return DataFrame with quality level information
        """
        return self.cdc.get_dataframe(path='/help/PH_Beschreibung_Phaenologie_Qualitaetsniveau.txt', index_column=0)

    def get_quality_bytes(self):
        """
        Return DataFrame with quality bytes information
        ftp://opendata.dwd.de/climate_environment/CDC/observations_germany/climate/subdaily/standard_format/qualitaetsbytes.pdf
        """
        return self.cdc.get_dataframe(path='/help/PH_Beschreibung_Phaenologie_Qualitaetsbyte.txt', index_column=0)

    def get_stations(self, filter=None, all=False):
        """
        Return DataFrame with stations information.
        """

        if self.dataset == 'immediate':
            filename = '/help/PH_Beschreibung_Phaenologie_Stationen_Sofortmelder.txt'
        elif self.dataset == 'annual':
            filename = '/help/PH_Beschreibung_Phaenologie_Stationen_Jahresmelder.txt'
        else:
            raise KeyError('Unknown dataset "{}"'.format(self.dataset))

        # Read stations CSV file
        data = self.cdc.get_dataframe(path=filename, index_column=0)

        # Unless "all==True", use only rows with "Datum Stationsaufloesung" == nan
        if not all:
            data = data[data['Datum Stationsaufloesung'].isna()]

        # Appropriately coerce geolocation values to float
        #dataframe_coerce_columns(data, ['geograph.Breite', 'geograph.Laenge'], float)

        # A. Basic filtering

        if filter:
            # Build "boolean indexing" filter expression from multiple ID-based criteria
            # https://pandas.pydata.org/pandas-docs/stable/indexing.html#boolean-indexing
            reference_fields = [
                'Stationsname',
                'Naturraumgruppe',
                'Naturraum',
                'Bundesland',
            ]

            expression = False
            for reference_field in reference_fields:
                #print(reference_field, filter)
                expression |= data[reference_field].str.contains(filter, case=False)

            # Apply filter expression to DataFrame
            if type(expression) is not bool:
                data = data[expression]

        return data

    def nearest_station(self, latitude, longitude, all=False):
        """
        Select closest station.
        """
        return self.nearest_stations(latitude, longitude, all=all).head(1)

    def nearest_stations(self, latitude, longitude, all=False, limit=10):
        """
        Select closest stations.

        Stolen from https://github.com/marians/dwd-weather
        """

        # Retrieve stations
        stations = self.get_stations(all=all)

        # Build list of distances to corresponding station
        distances = []
        for index, station in stations.iterrows():
            distance = haversine_distance(
                (longitude, latitude),
                (station['geograph.Laenge'], station['geograph.Breite'])
            )
            distances.append(distance)

        # Insert list of distances as new column
        stations.insert(1, 'Distanz', distances)

        # Sort ascending by distance value and limit row count
        stations = stations.sort_values('Distanz').head(n=limit)

        # Convert Series to DataFrame again
        frame = pd.DataFrame(stations)

        return frame

    def get_observations(self, options, humanize=False):
        """
        Retrieve observations.

        - Obtain query options
        - Compute DataFrame with combined observation data
        - Apply a bunch of filters to the result data
        """

        # Acquire data
        observations = self.query(partition=options['partition'], files=options.get('filename'))

        # Sanity checks
        if observations is None:
            return

        # Filter data
        observations = self.flux(observations, criteria=options)

        # Sanity checks
        if observations is None:
            logger.warning('No results found')
            return

        # Optionally humanize DataFrame
        if humanize:
            megaframe = self.create_megaframe(observations)
            observations = self.humanizer.get_observations(megaframe)

        # or pass-through with minor cosmetic amendments
        else:
            observations['Eintrittsdatum'] = observations['Eintrittsdatum'].astype(str)

        return observations

    def get_forecast(self, options, forecast_year=None, humanize=False):
        """
        Forecast observations.

        - Obtain query options
        - Get real observations, all filtering options can be used
        - Group results by (Stations_id, Objekt_id, Phase_id)
        - Aggregate mean "day of the year" value of the "Jultag" values for each group
        """

        # Compute target year.
        target_year = datetime.today().year
        if forecast_year is not None:
            target_year = datetime.strptime(str(forecast_year), "%Y").year

        # Get current observations
        observations = self.get_observations(options)

        # Sanity checks
        if observations is None:
            return

        # Group by station, species and phase
        # https://pandas.pydata.org/pandas-docs/stable/groupby.html
        grouped = observations.groupby(['Stations_id', 'Objekt_id', 'Phase_id'])

        # Aggregate mean "day of the year" value of the "Jultag" values for each group
        series = grouped['Jultag'].mean().round().astype(int)

        # Convert Series to DataFrame
        forecast = series.to_frame()


        # Compute ISO date from "day of the year" values and insert as new column
        real_dates = pd.to_datetime(target_year * 1000 + forecast['Jultag'], errors='coerce', format='%Y%j')
        forecast.insert(0, 'Datum', real_dates)

        # Resolve index column to real columns
        station_ids = []
        species_ids = []
        phase_ids = []
        for index, row in forecast.iterrows():
            # Decode index column and collect its components
            station_id, species_id, phase_id = index
            station_ids.append(station_id)
            species_ids.append(species_id)
            phase_ids.append(phase_id)

        forecast['Stations_id'], forecast['Objekt_id'], forecast['Phase_id'] = station_ids, species_ids, phase_ids


        # Optionally humanize DataFrame
        if humanize:
            megaframe = self.create_megaframe(forecast, drop_index_columns=True)
            forecast = self.humanizer.get_forecast(megaframe, target_year=target_year)

        # or pass-through with minor cosmetic amendments
        else:
            forecast['Datum'] = forecast['Datum'].astype(str)

        return forecast

    def query(self, partition=None, files=None):
        """
        The FTP/pandas workhorse, converges data from multiple observation data
        CSV files on upstream CDC FTP server into a single pandas DataFrame object.

        - Obtains ``partition`` parameter which can be either ``annual`` or ``immediate``.
        - Obtains optional ``files`` parameter which will be applied
          as an "include" filter to the list of scanned file names.
        """

        logger.info('Scanning for files')

        # Search FTP server
        paths = self.scan_files(partition, include=files, field='url')

        logger.info('Starting data acquisition with {} files'.format(len(paths)))

        # The main DataFrame object
        results = pd.DataFrame()

        # Load multiple files into single DataFrame
        for path in iterate_with_progressbar(paths):

            logger.debug('Processing file "{}"'.format(path))

            # Acquire DataFrame from CSV data
            data = self.cdc.get_dataframe(path, coerce_int=True)

            # Sanity checks
            if data is None:
                logger.warning('File "{}" is empty')
                continue

            # Coerce "Eintrittsdatum" column into date format
            data['Eintrittsdatum'] = pd.to_datetime(data['Eintrittsdatum'], errors='coerce', format='%Y%m%d')

            results = pd.concat([results, data], sort=False)

        # Sanity checks
        if results.empty:
            logger.info('Querying DWD CDC returned empty results')
            return

        # Reset index column
        results.reset_index(drop=True, inplace=True)

        return results

    def create_megaframe(self, frame, drop_index_columns=False):

        # https://pandas.pydata.org/pandas-docs/stable/merging.html#database-style-dataframe-joining-merging

        # Prevent errors like::
        #   ValueError: 'Stations_id' is both an index level and a column label, which is ambiguous.
        if drop_index_columns:
            frame.drop(columns=['Stations_id'], axis='columns', inplace=True)

        # Stations
        frame = pd.merge(frame, self.get_stations(), left_on='Stations_id', right_on='Stations_id')

        # Species
        frame = pd.merge(frame, self.get_species(), left_on='Objekt_id', right_on='Objekt_ID')

        # Phases
        frame = pd.merge(frame, self.get_phases(), left_on='Phase_id', right_on='Phase_ID')

        # Quality level
        if 'Qualitaetsniveau' in frame:
            frame = pd.merge(frame, self.get_quality_levels(), left_on='Qualitaetsniveau', right_index=True)

        # Quality byte
        # ftp://opendata.dwd.de/climate_environment/CDC/observations_germany/climate/subdaily/standard_format/qualitaetsbytes.pdf
        if 'Eintrittsdatum_QB' in frame:
            frame = pd.merge(frame, self.get_quality_bytes(), left_on='Eintrittsdatum_QB', right_index=True)

        # Debugging
        #print frame.to_csv(encoding='utf-8')

        return frame

    def flux(self, results, criteria=None):
        """
        The flux compensator. All filtering on the DataFrame takes places here.
        """

        logger.info('Entering flux compensator: Filter and transform data')

        criteria = criteria or {}

        # Sanity checks
        if results is None:
            return

        # Filter DataFrame
        # https://pythonspot.com/pandas-filter/
        # https://stackoverflow.com/questions/12065885/filter-dataframe-rows-if-value-in-column-is-in-a-set-list-of-values/12065904#12065904


        # A. Basic filtering

        # Build "boolean indexing" filter expression from multiple ID-based criteria
        # https://pandas.pydata.org/pandas-docs/stable/indexing.html#boolean-indexing
        isin_map = {
            'year': 'Referenzjahr',
            'quality-level': 'Qualitaetsniveau',
            'quality-byte': 'Eintrittsdatum_QB',
            'station-id': 'Stations_id',
            'species-id': 'Objekt_id',
            'phase-id': 'Phase_id',
        }

        # Low-level filtering based on IDs
        # For each designated field, add ``.isin`` criteria to "boolean index" expression
        expression = True
        for key, field in list(isin_map.items()):
            if field not in results:
                continue
            reference = results[field]
            if key in criteria and criteria[key]:
                values = map(int, criteria[key])
                expression &= reference.isin(values)

        # Apply filter expression to DataFrame
        if type(expression) is not bool:
            results = results[expression]


        # B. Humanized filtering based on merged/joined DataFrames

        # Build "boolean indexing" filter expression from multiple text-based criteria
        # https://pandas.pydata.org/pandas-docs/stable/indexing.html#boolean-indexing
        # https://stackoverflow.com/questions/12065885/filter-dataframe-rows-if-value-in-column-is-in-a-set-list-of-values/26724725#26724725
        patterns_map = {
            'station': ['Stationsname', 'Naturraumgruppe', 'Naturraum', 'Bundesland'],
            'species': ['Objekt', 'Objekt_englisch', 'Objekt_latein'],
            'phase':   ['Phase', 'Phase_englisch'],
            'quality': ['Beschreibung_x', 'Beschreibung_y'],
        }

        # Remember columns of current DataFrame
        columns = results.columns

        expression = True
        is_megaframe = False
        for field, reference_fields in list(patterns_map.items()):

            if field in criteria and criteria[field]:

                # Create megaframe as this will contain all information required for advanced searching
                if not is_megaframe:
                    is_megaframe = True
                    results = self.create_megaframe(results)

                # The list of patterns to search for. Any match counts.
                pattern = '|'.join(criteria[field])

                subexpression = False
                for reference_field in reference_fields:
                    subexpression |= results[reference_field].str.contains(pattern, case=False)

                expression &= subexpression

        # Shrink megaframe back using projection on former columns
        results = results[columns]

        # Apply filter expression to DataFrame
        if type(expression) is not bool:
            results = results[expression]

        return results

    def scan_files(self, partition, include=None, field=None):
        """
        Scan upstream files in three-level directory hierarchy.
        """

        # The full URL to the FTP data directory
        url = self.cdc.baseurl + self.data_directory

        # Query FTP server for files
        entries = self.cdc.ftp.scan_files(
            url, subdir=partition,
            include=include,
            include_base=['PH_(Sofort|Jahres)melder.+\.txt'],
            exclude_base=['PH_Beschreibung', 'Spezifizierung', 'PH_.+_Notiz'],
        )

        # Return entries if projection to field not requested
        if not field:
            return entries

        # Project entries to results
        results = []
        for entry in entries:

            try:
                item = entry[field]
                results.append(item)

            except KeyError:
                raise KeyError('Projection "field={}" not available'.format(field))

        return results

    @classmethod
    def load_preset(cls, section, option, name):
        resource = pkg_resources.resource_stream(__name__, 'presets.json')
        presets = json.load(resource)
        try:
            value = presets[section][option][name]
            return value
        except KeyError:
            message = 'Preset "{}" not found in file "{}"'.format(name, resource.name)
            logger.error(message)
            raise KeyError(message)

@attr.s
class DwdPhenoDataHumanizer:
    """
    Bring result DataFrame in a more pleasant shape.
    """

    # Output language. "english", "german" or "latin"
    language = attr.ib()

    # Whether to output long station name including "Naturraumgruppe" and "Naturraum"
    long_station = attr.ib()

    # Whether to show IDs alongside resolved text representation
    show_ids = attr.ib()

    def get_observations(self, frame):

        canvas = pd.DataFrame()

        stations, species, phases, quality_levels, quality_bytes = self.get_fields(frame)

        # Build fresh DataFrame with designated order of columns
        canvas['Jahr'] = frame['Referenzjahr']
        canvas['Datum'] = frame['Eintrittsdatum'].astype(str)
        canvas['Tag'] = frame['Jultag']
        canvas['Spezies'] = species
        canvas['Phase'] = phases
        canvas['Station'] = stations
        canvas['QS-Level'] = quality_levels
        canvas['QS-Byte'] = quality_bytes

        return canvas

    def get_forecast(self, frame, target_year):

        canvas = pd.DataFrame()

        stations, species, phases, quality_levels, quality_bytes = self.get_fields(frame)

        # Build fresh DataFrame with designated order of columns
        canvas['Datum'] = frame['Datum'].astype(str)
        canvas['Tag'] = frame['Jultag']
        canvas['Spezies'] = species
        canvas['Phase'] = phases
        canvas['Station'] = stations

        canvas.insert(0, 'Jahr', target_year)

        return canvas

    def get_fields(self, frame):

        # Which fields to use from "station" entity
        if self.long_station:
            station_fields = ['Stationsname', 'Naturraumgruppe', 'Naturraum', 'Bundesland']
        else:
            station_fields = ['Stationsname', 'Bundesland']

        # Improved map for quality level texts
        quality_level_text = {
            1: u'Load time checks',
            7: u'ROUTKLI validated',
            10: u'ROUTKLI validated and corrected',
            }

        # Which field to choose from the "species" entity. One of "Objekt", "Objekt_englisch", "Objekt_latein".
        # Which field to choose from the "phase" entity. One of "Phase", "Phase_englisch".
        species_field = 'Objekt_englisch'
        phase_field = 'Phase_englisch'
        if self.language:
            language = self.language.lower()
            if language == 'german':
                species_field = 'Objekt'
                phase_field = 'Phase'
                quality_level_text = {
                    1: u'Vorabprüfung beim Laden',
                    7: u'ROUTKLI geprüft',
                    10: u'ROUTKLI geprüft und korrigiert',
                    }
            elif language == 'latin':
                species_field = 'Objekt_latein'

        stations = []
        species = []
        phases = []
        quality_levels = []
        quality_bytes = []
        for index, row in frame.iterrows():

            # Station
            station_parts = [row[field] for field in station_fields if field in row]
            station_label = ', '.join(station_parts)
            if self.show_ids:
                station_label += ' [{}]'.format(row['Stations_id'])
            stations.append(station_label)

            # Species
            species_label = row.get(species_field, '')
            if self.show_ids:
                species_label += ' [{}]'.format(row['Objekt_id'])
            species.append(species_label)

            # Phase
            phase_label = row.get(phase_field, '')
            if self.show_ids:
                phase_label += ' [{}]'.format(row['Phase_id'])
            phases.append(phase_label)

            # Qualitaetsniveau
            if 'Qualitaetsniveau' in row:
                ql_label = quality_level_text.get(row['Qualitaetsniveau'], row.get('Beschreibung_x', ''))
                if self.show_ids:
                    ql_label += ' [{}]'.format(row['Qualitaetsniveau'])
                quality_levels.append(ql_label)

            # Qualitaetsbyte
            # ftp://opendata.dwd.de/climate_environment/CDC/observations_germany/climate/subdaily/standard_format/qualitaetsbytes.pdf
            if 'Eintrittsdatum_QB' in row:
                qb_label = row.get('Beschreibung_y', '')
                if self.show_ids:
                    qb_label += ' [{}]'.format(row['Eintrittsdatum_QB'])
                quality_bytes.append(qb_label)

        return stations, species, phases, quality_levels, quality_bytes



import json

import marko
from datadiff.tools import assert_equal

from tests.util import run_command




def test_cli_observations_immediate_recent_filter_station_id(capsys):
    """
    CLI test: Verify the `observations` subcommand works.
    """
    run_command("phenodata observations --source=dwd --dataset=immediate --partition=recent --filename=Hasel --station-id=19475 --humanize --show-ids --format=json")

    out, err = capsys.readouterr()
    response = json.loads(out)

    first = {
        "Jahr": 2021,
        "Datum": "2021-02-11",
        "Tag": 42,
        "Spezies": "common hazel [113]",
        "Phase": "beginning of flowering [5]",
        "Station": "Angermünde (Ph), Brandenburg [19475]",
        "QS-Level": "ROUTKLI validated [7]",
        "QS-Byte": "Feldwert nicht beanstandet [1]"
    }
    assert_equal(response[0], first)


def test_cli_observations_immediate_recent_filter_station_name(capsys):
    """
    CLI test: Verify the `observations` subcommand works.
    """
    run_command("phenodata observations --source=dwd --dataset=annual --partition=recent --filename=Hasel --station=berlin,brandenburg --humanize --sort=Datum --format=json")

    out, err = capsys.readouterr()
    response = json.loads(out)

    first = {
        "Jahr": 2021,
        "Datum": "2021-02-01",
        "Tag": 32,
        "Spezies": "common hazel",
        "Phase": "beginning of flowering",
        "Station": "Wall, Brandenburg",
        "QS-Level": "ROUTKLI validated and corrected",
        "QS-Byte": "Feldwert nicht beanstandet"
    }
    assert_equal(response[0], first)


def test_cli_observations_immediate_historical(capsys):
    """
    CLI test: Verify the `observations` subcommand works.
    """
    run_command("phenodata observations --source=dwd --dataset=immediate --partition=historical --filename=Hasel --station=berlin,brandenburg --humanize --sort=Datum --format=json")

    out, err = capsys.readouterr()
    response = json.loads(out)

    first = {
        "Jahr": 2007,
        "Datum": "2007-01-12",
        "Tag": 12,
        "Spezies": "common hazel",
        "Phase": "beginning of flowering",
        "Station": "Prenzlau, Brandenburg",
        "QS-Level": "ROUTKLI validated",
        "QS-Byte": "Feldwert nicht beanstandet"
    }
    assert_equal(response[0], first)


def test_cli_observations_annual_recent(capsys):
    """
    CLI test: Verify the `observations` subcommand works.
    """
    run_command("phenodata observations --source=dwd --dataset=annual --partition=recent --filename=Hasel --station=berlin,brandenburg --humanize --sort=Datum --format=json")

    out, err = capsys.readouterr()
    response = json.loads(out)

    first = {
        "Jahr": 2021,
        "Datum": "2021-02-01",
        "Tag": 32,
        "Spezies": "common hazel",
        "Phase": "beginning of flowering",
        "Station": "Wall, Brandenburg",
        "QS-Level": "ROUTKLI validated and corrected",
        "QS-Byte": "Feldwert nicht beanstandet"
    }
    assert_equal(response[0], first)


def test_cli_observations_annual_historical(capsys):
    """
    CLI test: Verify the `observations` subcommand works.
    """
    run_command("phenodata observations --source=dwd --dataset=annual --partition=historical --filename=Hasel --station=berlin,brandenburg --humanize --sort=Datum --format=json")

    out, err = capsys.readouterr()
    response = json.loads(out)

    first = {
        "Jahr": 1936,
        "Datum": "1936-03-10",
        "Tag": 70,
        "Spezies": "common hazel",
        "Phase": "beginning of flowering",
        "Station": "Berlin-Dahlem, Berlin",
        "QS-Level": "Load time checks",
        "QS-Byte": "Feldwert nicht beanstandet"
    }
    assert_equal(response[0], first)


def test_cli_observations_filter_year(capsys):
    """
    CLI test: Verify the `observations` subcommand works, with filtering by year.
    """
    run_command("phenodata observations --source=dwd --dataset=immediate --partition=recent --filename=Hasel --station-id=7521,7532 --year=2020,2021 --humanize --show-ids --format=json")

    out, err = capsys.readouterr()
    response = json.loads(out)

    first = {
        "Jahr": 2021,
        "Datum": "2021-02-24",
        "Tag": 55,
        "Spezies": "common hazel [113]",
        "Phase": "beginning of flowering [5]",
        "Station": "Norder-Hever-Koog, Schleswig-Holstein [7532]",
        "QS-Level": "ROUTKLI validated [7]",
        "QS-Byte": "Feldwert nicht beanstandet [1]"
    }
    assert_equal(response[0], first)


def test_cli_observations_filter_species_id(capsys):
    """
    CLI test: Verify the `observations` subcommand works, with filtering by species-id.
    """
    run_command("phenodata observations --source=dwd --dataset=immediate --partition=recent --filename=Hasel --station-id=7521,7532 --species-id=113,127 --humanize --show-ids --format=json")

    out, err = capsys.readouterr()
    response = json.loads(out)

    first = {
        "Jahr": 2021,
        "Datum": "2021-02-24",
        "Tag": 55,
        "Spezies": "common hazel [113]",
        "Phase": "beginning of flowering [5]",
        "Station": "Norder-Hever-Koog, Schleswig-Holstein [7532]",
        "QS-Level": "ROUTKLI validated [7]",
        "QS-Byte": "Feldwert nicht beanstandet [1]"
    }
    assert_equal(response[0], first)


def test_cli_observations_filter_invalid_readings(capsys):
    """
    CLI test: Verify the `observations` subcommand works, with filtering by quality-byte.
    """
    run_command("phenodata observations --source=dwd --dataset=immediate --partition=recent --filename=Hasel --quality-byte=5,6,7,8 --humanize --show-ids --format=json")

    out, err = capsys.readouterr()
    response = json.loads(out)

    first = {
        "Jahr": 2021,
        "Datum": "2021-02-20",
        "Tag": 51,
        "Spezies": "common hazel [113]",
        "Phase": "beginning of flowering [5]",
        "Station": "Kirchdorf b. Sulingen, Niedersachsen [7857]",
        "QS-Level": "ROUTKLI validated [7]",
        "QS-Byte": "Feldwert zweifelhaft [5]"
    }
    assert_equal(response[0], first)


def test_cli_observations_filter_sql(capsys):
    """
    CLI test: Verify the `observations` subcommand works, with filtering by quality-byte.
    """
    run_command("""
    phenodata observations \
        --source=dwd --dataset=annual --partition=recent \
        --filename=Hasel \
        --year=2022 \
        --species-preset=mellifera-de-primary --phase="beginning of flowering" \
        --humanize --language=german \
        --sql="SELECT * FROM data WHERE Station LIKE '%Berlin%' ORDER BY Datum" \
        --format=json
    """)

    out, err = capsys.readouterr()
    response = json.loads(out)

    first = {
        "Jahr": 2022,
        "Datum": "2022-01-11",
        "Tag": 11,
        "Spezies": "Hasel",
        "Phase": "Blüte Beginn",
        "Station": "Berlin-Marienfelde, Berlin",
        "QS-Level": "ROUTKLI geprüft und korrigiert",
        "QS-Byte": "Feldwert nicht beanstandet"
    }
    assert_equal(response[0], first)


def test_cli_observations_format_csv(capsys):
    """
    CLI test: Verify the `observations` subcommand works with CSV output.
    """
    run_command("phenodata observations --source=dwd --dataset=immediate --partition=recent --filename=Hasel --station-id=19475 --humanize --show-ids --format=csv")

    out, err = capsys.readouterr()
    assert out.startswith("""
Jahr,Datum,Tag,Spezies,Phase,Station,QS-Level,QS-Byte
2021,2021-02-11,42,common hazel [113],beginning of flowering [5],"Angermünde (Ph), Brandenburg [19475]",ROUTKLI validated [7],Feldwert nicht beanstandet [1]
2022,2022-01-28,28,common hazel [113],beginning of flowering [5],"Angermünde (Ph), Brandenburg [19475]",Load time checks [1],Feldwert nicht beanstandet [1]
    """.strip())


def test_cli_observations_format_tabular(capsys):
    """
    CLI test: Verify the `observations` subcommand works with tabular output.

    `tabular:pipe` actually yields a Markdown table, so let's validate it using a Markdown parser.
    """
    run_command("phenodata observations --source=dwd --dataset=immediate --partition=recent --filename=Hasel --station-id=19475 --humanize --show-ids --format=tabular:pipe")

    out, err = capsys.readouterr()

    html = marko.convert(out)

    assert html.startswith("<p>|   Jahr | Datum      |   Tag | Spezies")
    assert html.endswith("| Feldwert nicht beanstandet [1] |</p>\n")



# -*- coding: utf-8 -*-
# (c) 2018-2023, The Earth Observations Developers
"""
phenodata is an acquisition and processing toolkit for open access phenology data.
"""
import sys
import logging
from docopt import docopt, DocoptExit
from tabulate import tabulate
from phenodata import __version__
from phenodata.ftp import FTPSession
from phenodata.dwd.cdc import DwdCdcClient
from phenodata.dwd.pheno import DwdPhenoData, DwdPhenoDataHumanizer
from phenodata.util import boot_logging, normalize_options, options_convert_lists

logger = logging.getLogger(__name__)

APP_NAME = 'phenodata'

def run():
    """
    Usage:
      phenodata info
      phenodata list-species --source=dwd [--format=csv]
      phenodata list-phases --source=dwd [--format=csv]
      phenodata list-stations --source=dwd --dataset=immediate [--all] [--filter=berlin] [--sort=Stationsname] [--format=csv]
      phenodata nearest-station --source=dwd --dataset=immediate --latitude=52.520007 --longitude=13.404954 [--format=csv]
      phenodata nearest-stations --source=dwd --dataset=immediate --latitude=52.520007 --longitude=13.404954 [--all] [--limit=10] [--format=csv]
      phenodata list-quality-levels --source=dwd [--format=csv]
      phenodata list-quality-bytes --source=dwd [--format=csv]
      phenodata list-filenames --source=dwd --dataset=immediate --partition=recent [--filename=Hasel,Schneegloeckchen] [--year=2017]
      phenodata list-urls --source=dwd --dataset=immediate --partition=recent [--filename=Hasel,Schneegloeckchen] [--year=2017]
      phenodata (observations|forecast) --source=dwd --dataset=immediate --partition=recent [--filename=Hasel,Schneegloeckchen] [--station-id=164,717] [--species-id=113,127] [--phase-id=5] [--quality-level=10] [--quality-byte=1,2,3] [--station=berlin,brandenburg] [--species=hazel,snowdrop] [--species-preset=mellifera-de-primary] [--phase=flowering] [--quality=ROUTKLI] [--year=2017] [--forecast-year=2021] [--humanize] [--show-ids] [--language=german] [--long-station] [--sort=Datum] [--sql=sql] [--format=csv] [--verbose]
      phenodata drop-cache --source=dwd
      phenodata --version
      phenodata (-h | --help)

    Data acquisition options:
      --source=<source>         Data source. Currently, only "dwd" is a valid identifier.
      --dataset=<dataset>       Data set. Use "immediate" or "annual" for "--source=dwd".
      --partition=<dataset>     Partition. Use "recent" or "historical" for "--source=dwd".
      --filename=<file>         Filter by file names (comma-separated list)

    Direct filtering options:
      --year=<year>             Filter by year (comma-separated list)
      --station-id=<station-id> Filter by station identifiers (comma-separated list)
      --species-id=<species-id> Filter by species identifiers (comma-separated list)
      --phase-id=<phase-id>     Filter by phase identifiers (comma-separated list)

    Humanized filtering options:
      --station=<station>       Filter by strings from "stations" data (comma-separated list)
      --species=<species>       Filter by strings from "species" data (comma-separated list)
      --phase=<phase>           Filter by strings from "phases" data (comma-separated list)
      --species-preset=<preset> Filter by strings from "species" data (comma-separated list)
                                The preset will get loaded from the "presets.json" file.

    Forecasting options:
      --forecast-year=<year>    Use as designated forecast year.

    Postprocess filtering options:
      --sql=<sql>               Apply given SQL query before output.

    Data output options:
      --format=<format>         Output data in designated format. Choose one of "tabular", "json",
                                "csv", or "string". With "tabular", it is also possible to specify
                                the table format. Use "tabular:pipe" for Markdown output, or
                                "tabular:rst" for reStructuredText. [default: tabular:psql]
      --sort=<sort>             Sort by given field names. (comma-separated list)
      --humanize                Resolve identifier-based fields to human-readable labels.
      --show-ids                Show identifiers alongside resolved labels, when using "--humanize".
      --language=<language>     Use labels in designated language, when using "--humanize"
                                [default: english].
      --long-station            Use long station name including "Naturraumgruppe" and "Naturraum".
      --limit=<limit>           Limit output of "nearest-stations" to designated number of entries.
                                [default: 10]
      --verbose                 Turn on verbose output.
    """

    # Use generic commandline options schema and amend with current program name
    commandline_schema = run.__doc__

    # Read commandline options
    options = docopt(commandline_schema, version=APP_NAME + ' ' + __version__)

    # Initialize logging
    boot_logging(options)

    # Normalize commandline options
    options = normalize_options(options, encoding='utf-8')

    # Expand options
    preset_name = options['species-preset']
    if preset_name:
        options['species'] = DwdPhenoData.load_preset('options', 'species', preset_name)

    # Coerce comma-separated list fields
    options_convert_lists(options, list_items=[

        # Acquisition parameters
        'filename',

        # Filter parameters
        'year',

        # ID parameters
        'quality-level',
        'quality-byte',
        'station-id',
        'species-id',
        'phase-id',

        # Humanized parameters
        'quality',
        'station',
        'species',
        'phase',

        # Sorting parameters
        'sort',
    ])

    # Command line argument debugging
    #import pprint; print 'options:\n{}'.format(pprint.pformat(options))

    if options['info']:
        print('Name:         phenodata-{version}'.format(version=__version__))
        print('Description:  phenodata is an acquisition and processing toolkit for open access phenology data')
        print('Data sources: DWD')
        # TODO: Add cache location and info
        return

    # Create data source adapter
    if options['source'] == 'dwd':
        cdc_client = DwdCdcClient(ftp=FTPSession())
        humanizer = DwdPhenoDataHumanizer(language=options['language'], long_station=options['long-station'], show_ids=options['show-ids'])
        client = DwdPhenoData(cdc=cdc_client, humanizer=humanizer, dataset=options.get('dataset'))
    else:
        message = 'Data source "{}" not implemented'.format(options['source'])
        logger.error(message)
        raise DocoptExit(message)

    # Dispatch command
    data = None
    if options['list-species']:
        data = client.get_species()
    elif options['list-phases']:
        data = client.get_phases()
    elif options['list-stations']:
        data = client.get_stations(filter=options['filter'], all=options['all'])
    elif options['list-quality-levels']:
        data = client.get_quality_levels()
    elif options['list-quality-bytes']:
        data = client.get_quality_bytes()

    elif options['list-filenames']:
        files = client.scan_files(options['partition'], include=options['filename'], field='name')
        print('\n'.join(files))
        return
    elif options['list-urls']:
        files = client.scan_files(options['partition'], include=options['filename'], field='url')
        print('\n'.join(files))
        return

    elif options['observations']:
        data = client.get_observations(options, humanize=options['humanize'])

    elif options['forecast']:
        data = client.get_forecast(options, forecast_year=options['forecast-year'], humanize=options['humanize'])

    elif options['nearest-station']:
        data = client.nearest_station(float(options['latitude']), float(options['longitude']), all=options['all'])

    elif options['nearest-stations']:
        data = client.nearest_stations(float(options['latitude']), float(options['longitude']), all=options['all'], limit=int(options['limit']))

    elif options['drop-cache']:
        client.cdc.ftp.ensure_cache_manager()
        if client.cdc.ftp.cache.drop():
            logger.info('Dropping the cache succeeded')
        else:
            logger.warning('Dropping the cache failed')
        return

    # Query results
    if data is not None and options["sql"]:
        import duckdb
        data = duckdb.query(query=options["sql"], alias="data").df()

    # Format and output results
    if data is not None:

        output_format = options['format'].lower()

        # Whether to show the index column or not
        showindex = True
        if options['observations'] or options['forecast']:
            showindex = False

        # Sort columns
        if options['sort']:
            data.sort_values(options['sort'], inplace=True)

        # Handle aliases for reStructuredText and Markdown
        if output_format in ["restructuredtext", "rst"]:
            output_format = "tabular:rst"
        elif output_format in ["markdown", "md"]:
            output_format = "tabular:pipe"

        output = None
        if output_format.startswith('tabular'):

            try:
                tablefmt = output_format.split(':')[1]
            except:
                tablefmt = 'psql'

            # TODO: How to make "tabulate" print index column name?
            output = tabulate(data, headers=data.columns, showindex=showindex, tablefmt=tablefmt)

        elif output_format == 'csv':
            output = data.to_csv(encoding='utf-8', index=showindex)

        elif output_format == 'json':
            if showindex:
                data = data.reset_index()
            output = data.to_json(orient='records', date_format='iso')

        elif output_format == 'string':
            output = data.to_string()

        else:
            message = 'Unknown output format "{}"'.format(options['format'])
            logger.error(message)
            sys.exit(1)

        if output is not None:
            if sys.version_info.major == 2:
                print(output.encode('utf-8'))
            else:
                print(output)
        else:
            logger.warning('Empty output')



import json
import sys

import pytest
from datadiff.tools import assert_equal

from tests.util import run_command


def test_cli_list_species(capsys):
    """
    CLI test: Verify the `list-species` subcommand works.
    """
    run_command("phenodata list-species --source=dwd --format=json")

    out, err = capsys.readouterr()
    response = json.loads(out)

    first = {
      "Objekt_ID": 25,
      "Objekt": "Rüben",
      "Objekt_englisch": "beet",
      "Objekt_latein": "Beta vulgaris"
    }
    assert_equal(response[0], first)


def test_cli_list_phases(capsys):
    """
    CLI test: Verify the `list-phases` subcommand works.
    """
    run_command("phenodata list-phases --source=dwd --format=json")

    out, err = capsys.readouterr()
    response = json.loads(out)

    first = {
      "Phase_ID": 1,
      "Phase": "Ergrünen Beginn",
      "Phase_englisch": "beginning of turning green"
    }
    assert_equal(response[0], first)


def test_cli_list_quality_levels(capsys):
    """
    CLI test: Verify the `list-quality-levels` subcommand works.
    """
    run_command("phenodata list-quality-levels --source=dwd --format=json")

    out, err = capsys.readouterr()
    response = json.loads(out)

    reference = [
      {
        "Qualitaetsniveau": 1,
        "Beschreibung": "nur formale Prüfung beim Entschlüsseln und Laden"
      },
      {
        "Qualitaetsniveau": 7,
        "Beschreibung": "in ROUTINE geprüft, aber keine Korrekturen (z.B. RR_UN vor Korrektur)"
      },
      {
        "Qualitaetsniveau": 10,
        "Beschreibung": "in ROUTINE geprüft, routinemäßige Korrektur beendet"
      }
    ]
    assert_equal(response, reference)


@pytest.mark.skipif(sys.platform == "linux", reason="Charset encoding weirdness. Works on macOS.")
def test_cli_list_quality_bytes(capsys):
    """
    CLI test: Verify the `list-quality-bytes` subcommand works.
    """
    run_command("phenodata list-quality-bytes --source=dwd --format=json")

    out, err = capsys.readouterr()
    response = json.loads(out)

    reference = [
      {
        "Qualiaetsbyte": 0,
        "Beschreibung": "Feldwert ungeprüft"
      },
      {
        "Qualiaetsbyte": 1,
        "Beschreibung": "Feldwert nicht beanstandet"
      },
      {
        "Qualiaetsbyte": 2,
        "Beschreibung": "Feldwert korrigiert"
      },
      {
        "Qualiaetsbyte": 3,
        "Beschreibung": "Feldwert trotz Beanstandung bestätigt"
      },
      {
        "Qualiaetsbyte": 5,
        "Beschreibung": "Feldwert zweifelhaft"
      },
      {
        "Qualiaetsbyte": 7,
        "Beschreibung": "ungültiges Eintrittsdatum, z.B. 31. April, wird automatisch"
      },
      {
        "Qualiaetsbyte": 8,
        "Beschreibung": "Feldwert falsch"
      }
    ]
    assert_equal(response, reference)


def test_cli_list_filenames_immediate_recent(capsys):
    """
    CLI test: Verify the `list-filenames` subcommand works.
    """
    run_command("phenodata list-filenames --source=dwd --dataset=immediate --partition=recent")

    out, err = capsys.readouterr()
    response = out.splitlines()

    assert_equal(response[0], "PH_Sofortmelder_Landwirtschaft_Kulturpflanze_Dauergruenland_akt.txt")
    assert_equal(response[-1], "PH_Sofortmelder_Wildwachsende_Pflanze_Wiesen-Fuchsschwanz_akt.txt")


def test_cli_list_filenames_immediate_historical(capsys):
    """
    CLI test: Verify the `list-filenames` subcommand works.
    """
    run_command("phenodata list-filenames --source=dwd --dataset=immediate --partition=historical")

    out, err = capsys.readouterr()
    response = out.splitlines()

    assert_equal(response[0], "PH_Sofortmelder_Landwirtschaft_Kulturpflanze_Dauergruenland_1979_2021_hist.txt")
    assert_equal(response[-1], "PH_Sofortmelder_Wildwachsende_Pflanze_Wiesen-Fuchsschwanz_1979_2021_hist.txt")


def test_cli_list_filenames_annual_recent(capsys):
    """
    CLI test: Verify the `list-filenames` subcommand works.
    """
    run_command("phenodata list-filenames --source=dwd --dataset=annual --partition=recent")

    out, err = capsys.readouterr()
    response = out.splitlines()

    assert_equal(response[0], "PH_Jahresmelder_Landwirtschaft_Kulturpflanze_Dauergruenland_akt.txt")
    assert_equal(response[-1], "PH_Jahresmelder_Wildwachsende_Pflanze_Zweigriffliger_Weissdorn_akt.txt")


def test_cli_list_filenames_annual_historical(capsys):
    """
    CLI test: Verify the `list-filenames` subcommand works.
    """
    run_command("phenodata list-filenames --source=dwd --dataset=annual --partition=historical")

    out, err = capsys.readouterr()
    response = out.splitlines()

    assert_equal(response[0], "PH_Jahresmelder_Landwirtschaft_Kulturpflanze_Dauergruenland_1936_2021_hist.txt")
    assert_equal(response[-1], "PH_Jahresmelder_Wildwachsende_Pflanze_Zweigriffliger_Weissdorn_1936_2021_hist.txt")


def test_cli_list_urls_immediate_recent(capsys):
    """
    CLI test: Verify the `list-urls` subcommand works.
    """
    run_command("phenodata list-urls --source=dwd --dataset=immediate --partition=recent")

    out, err = capsys.readouterr()
    response = out.splitlines()

    assert response[0].startswith("ftp://opendata.dwd.de/climate_environment/CDC/observations_germany/phenology/immediate_reporters/crops/recent")



import json

from datadiff.tools import assert_equal

from tests.util import run_command




def test_cli_forecast_immediate_recent(capsys):
    """
    CLI test: Verify the `forecast` subcommand works.
    """
    run_command("phenodata forecast --source=dwd --dataset=immediate --partition=recent --filename=Hasel --station-id=7521,7532 --humanize --format=json")

    out, err = capsys.readouterr()
    response = json.loads(out)

    first = {
        "Jahr": 2023,
        "Datum": "2023-02-26",
        "Tag": 57,
        "Spezies": "common hazel",
        "Phase": "beginning of flowering",
        "Station": "Norder-Hever-Koog, Schleswig-Holstein"
    }
    assert_equal(response[0], first)


def test_cli_forecast_annual_recent(capsys):
    """
    CLI test: Verify the `forecast` subcommand works, also select German.

    Event sequence for each species
    -------------------------------
    Forecast of all observations at station "Berlin-Dahlem".
    Use all species of the "primary group" (dito).
    Sort by species and date, ascending.

    """
    run_command("""
    phenodata forecast \
        --source=dwd --dataset=annual --partition=recent \
        --filename=Apfel \
        --station-id=12132 \
        --humanize --language=german \
        --sort=Spezies,Datum \
        --format=json
    """)

    out, err = capsys.readouterr()
    response = json.loads(out)

    first = {
        "Jahr": 2023,
        "Datum": "2023-04-04",
        "Tag": 94,
        "Spezies": "Apfel, frühe Reife",
        "Phase": "Austrieb Beginn",
        "Station": "Berlin-Dahlem, Berlin"
    }
    assert_equal(response[0], first)



import os
import sys


def test_example_stations(capfd):
    os.system(f"{sys.executable} examples/stations.py")

    out, err = capfd.readouterr()
    assert "Index: 323 entries" in out
    assert "Data columns (total 10 columns)" in out
    assert "Stationsname" in out
    assert "[323 rows x 10 columns]" in out


def test_example_observations(capfd):
    os.system(f"{sys.executable} examples/observations.py")

    out, err = capfd.readouterr()
    assert "Index: 6 entries" in out
    assert "Data columns (total 8 columns)" in out
    assert "Stations_id" in out



import re

import pytest

from tests.util import run_command


def test_cli_info(capsys):
    """
    CLI test: Verify `phenodata info` works.
    """
    run_command("phenodata info")

    out, err = capsys.readouterr()
    assert "phenodata is an acquisition and processing toolkit" in out


def test_cli_version(capsys):
    """
    CLI test: Verify `phenodata info` works.
    """
    with pytest.raises(SystemExit):
        run_command("phenodata --version")

    out, err = capsys.readouterr()
    assert re.match("phenodata \d+\.\d+\.\d+.*", out)



import os
from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))
README = open(os.path.join(here, 'README.rst')).read()

requires = [
    'arrow>=0.12.1,<1.3',
    'attrs>=17.4.0',
    'docopt>=0.6.2',
    'dogpile.cache>=0.6.5,<2',
    'pandas>=1.3,<2.1',
    'platformdirs<4',
    'requests>=2.18.4,<3',
    'requests-ftp>=0.3.1,<4',
    'tabulate>=0.8.2,<0.10',
    'tqdm>=4.60,<5',
]

test_requires = [
    'datadiff>=2.0,<3',
    'marko<2',
    'proselint==0.13.1',
    'pytest>=6.1.0,<8',
    'pytest-cov<5',
    'pytest-doctest-ellipsis-markers',
]

setup(name='phenodata',
    version='0.13.1',
    description='phenodata is an acquisition and processing toolkit for open access phenology data',
    long_description=README,
    license="AGPL 3, EUPL 1.2",
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: End Users/Desktop",
        "Intended Audience :: Information Technology",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: European Union Public Licence 1.2 (EUPL 1.2)",
        "License :: OSI Approved :: GNU Affero General Public License v3",
        "Natural Language :: English",
        "Natural Language :: German",
        "Natural Language :: Latin",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: POSIX :: Linux",
        "Topic :: Database",
        "Topic :: Education",
        "Topic :: Internet :: File Transfer Protocol (FTP)",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Internet :: WWW/HTTP :: Indexing/Search",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Atmospheric Science",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Scientific/Engineering :: Visualization",
        "Topic :: Software Development :: Libraries",
        "Topic :: System :: Archiving",
        "Topic :: Text Processing :: Filters",
        "Topic :: Text Processing :: Indexing",
        "Topic :: Utilities",
        ],
    author='Andreas Motl',
    author_email='andreas.motl@panodata.org',
    url='https://github.com/earthobservations/phenodata',
    keywords='phenology phenology-data phenology-models phenology-modelling phenometrics '
             'opendata scientific research-data dwd gpm usa-npn',
    packages=find_packages(),
    include_package_data=True,
    package_data={
        'phenodata': [
            'dwd/*.json',
        ],
    },
    zip_safe=False,
    install_requires=requires,
    extras_require={
        'sql': ['duckdb>=0.3,<0.8'],
        'test': test_requires,
    },
    entry_points={
        'console_scripts': [
            'phenodata  = phenodata.command:run',
        ],
    },

)



# -*- coding: utf-8 -*-
# (c) 2018-2023, The Earth Observations Developers
import logging
import re
import attr
import pandas as pd
from six import StringIO
from phenodata.util import dataframe_strip_strings, dataframe_coerce_columns

logger = logging.getLogger(__name__)


@attr.s
class DwdCdcClient:
    """
    Base functionality for conveniently accessing the Climate Data Center (CDC)
    FTP server operated by »Deutscher Wetterdienst« (DWD).
    """

    # The base URL to data on the CDC FTP server
    baseurl = 'ftp://opendata.dwd.de/climate_environment/CDC'

    # Instance of ``phenodata.ftp.FTPSession`` object for lowlevel access to CDC FTP
    ftp = attr.ib()

    def get_dataframe(self, url=None, path=None, index_column=None, coerce_int=False):
        """
        Read single CSV file from FTP url and convert to pandas DataFrame object.

        Obtains either a full ``url`` parameter or a ``path`` parameter
        for addressing the remote resource. If the ``path`` parameter is given,
        ``self.baseurl`` will be prefixed to ``path`` before accessing the resource.

        Optionally obtains ``index_column`` parameter.
        Use this to set the index of designated index column.

        Optionally obtains ``coerce_int`` parameter.
        Use this to convert all values to integer format.
        """
        if path:
            url = self.baseurl + path
        logger.info("Retrieving resource {}".format(url))
        return self.csv_to_dataframe(self.read_csv(url), index_column=index_column, coerce_int=coerce_int)

    def read_csv(self, url):
        """
        Read CSV file from FTP url and apply response caching based on file modification time (mtime).
        Fixup different anomalies to make it compatible with ``pandas.read_csv``.
        """

        # Retrieve CSV file
        content = self.ftp.retr_cached(url, strip_base=self.baseurl)

        # Sanity checks
        if not content:
            return

        # Fix CSV formatting
        content = content.strip()
        content = content.replace('\r\n', '')
        content = re.sub(';eor;\s*', ';eor;\n', content)
        content = re.sub('; eor ;\s*', '; eor;\n', content)
        content = content.strip()

        # Fixups for specific fields
        if 'Qualitaetsbyte' in url:
            content = content.replace('Eintrittsdatum;', 'Eintrittsdatum,')

        # Fixups for specific files
        if 'Kulturpflanze_Ruebe_akt' in url or 'Kulturpflanze_Ruebe_hist' in url:
            fieldmap = {
                'STATIONS_ID': 'Stations_id',
                'OBJEKT_ID': 'Objekt_id',
                'PHASE_ID': 'Phase_id',
                'REFERENZJAHR': 'Referenzjahr',
                'EINTRITTSDATUM': 'Eintrittsdatum',
                'JULTAG': 'Jultag',
                'EINTRITTSDATUM_QB': 'Eintrittsdatum_QB',
                'QUALITAETSNIVEAU': 'Qualitaetsniveau',
            }
            for old, new in list(fieldmap.items()):
                content = content.replace(old, new)

        # Debugging
        #print('content:\n', content)

        return StringIO(content)

    def csv_to_dataframe(self, stream, index_column=None, coerce_int=False):
        """
        Read CSV data from stream into pandas DataFrame object.

        Optionally obtains ``index_column`` parameter.
        Use this to set the index of designated index column.

        Optionally obtains ``coerce_int`` parameter.
        Use this to convert all values to integer format.
        """

        # Sanity checks
        #if not stream or stream.len == 0:
        #    return

        # Read CSV into pandas DataFrame.
        # https://pandas.pydata.org/pandas-docs/stable/user_guide/io.html#csv-text-files
        df = pd.read_csv(
            stream, engine='c', encoding='utf-8',
            delimiter=';', skipinitialspace=True,
            skip_blank_lines=True,
            on_bad_lines="warn",
            low_memory=False
        )


        # A. Apply cleanups

        # Strip whitespace from all column headers
        # https://stackoverflow.com/questions/21606987/how-can-i-strip-the-whitespace-from-pandas-dataframe-headers/21607530#21607530
        df.rename(columns=lambda x: x.strip(), inplace=True)

        # Strip whitespace from all values
        df = df.apply(dataframe_strip_strings, axis=0)

        # Remove rows with empty values
        for sanitize_column in ['Eintrittsdatum', 'Jultag', 'eor']:
            if sanitize_column in df:
                df.dropna(subset=[sanitize_column], inplace=True)

        # Remove trailing nonsense column
        # FIXME: Only remove if there actually *are* nonsense columns
        last_column_index = len(df.columns) - 1
        last_column = df.columns[[last_column_index]]
        df.drop(last_column, axis=1, inplace=True)

        # Remove end-of-row tombstone
        df.drop('eor', axis=1, inplace=True)


        # B. Apply coercions

        # Coerce types into correct format
        if coerce_int:
            df = df.astype(int)

        # Initialize index column on DataFrame
        if index_column is not None:

            # Get name of index column
            index_column_name = df.columns[index_column]

            # Coerce values in index column to integer
            dataframe_coerce_columns(df, [index_column_name], int)

            # Establish index on index column
            df.set_index(index_column_name, inplace=True)

        return df
